/**
  OpenGL/GLUT/GLUI "playground" for 3D mesh-based physics.
  
  Dr. Orion Sky Lawlor, olawlor@acm.org, 2011-04-21 (Public Domain)
*/
#include "GL/glew.h" /* GL extension wrangler: http://glew.sourceforge.net/ */
#include "GL/glut.h" /* windows, mice, etc.  http://freeglut.sourceforge.net */
#include <stdlib.h> /* for "rand" */
#include <stdio.h> /* for printf */
#include "osl/vec4.h" /* 3D vector class */
#include "osl/mat4.h" /* for easy 4x4 matrix manipulation */
#include "GL/glui.h" /* user interface widgets: http://glui.sourceforge.net/ */

#include "ogl/glsl.h"
#include "ogl/glsl.cpp"

#include "glui/glui.h"

#include "ogl/minicam.h" /* ortho_frame, camera motion, key_down & events */

#include <vector> /* std::vector container */
#include <sstream> /* for istringstream */
#include <map>

#ifndef M_PI /* why doesn't windows math.h have this? */
#define M_PI 3.14159265358979
#endif

/* These variables are changed directly by the user interface code below */

/* Each display-time rendered layer has this alpha value */
float display_alpha=0.1;

/* Enable for anisotropic filtering */
int texaniso=1;

/* Texture pixel format on graphics card */
int texformat=GL_RGBA8; 

/* Texture "wrap mode" in X and Y: */
int wrap_mode=GL_CLAMP_TO_EDGE; // REPEAT;

/* Texture magnification mode/minification mode */
int minfilter_mode=GL_LINEAR_MIPMAP_LINEAR; 
int magfilter_mode=GL_LINEAR; 




/** This struct represents one uniform or constant variable in a GLSL program */
class GLSL_variable {
public:
	std::string name; /* name in GLSL code */
	int location; /* glGetUniformLocation of uniform */
	float value;  /* current value */
};

/** This class extracts variable variables from a GLSL shader,
and allows C++ to query them. */
class GLSL_variable_finder {
public:
	GLSL_variable_finder(std::string programText,GLUI *glui);
	
	/* Return the location of this GLSL uniform (or -1 if none exists) */
	int getUniformLocation(std::string variableName) {
		GLSL_variable *u=variables[variableName];
		if (u) return u->location;
		else return -1;
	}
	/* Return the value of this GLSL variable (or 0.0 if none exists) */
	float getVariableValue(std::string variableName) {
		GLSL_variable *u=variables[variableName];
		if (u) return u->value;
		else return 0.0;
	}
	
	/* Set the GLSL value of all our uniform variables */
	void setUniforms(GLhandleARB prog) {
		vars_t::iterator it=variables.begin();
		while (it!=variables.end()) {
			GLSL_variable *u=(*it).second;
			if (u->location==-2) u->location=glGetUniformLocationARB(prog,u->name.c_str());
			if (u->location>=0)
				glUniform1fARB(u->location,u->value);
			++it;
		}
	}
private:
	typedef std::map<std::string,GLSL_variable *> vars_t;
	vars_t variables;
};

GLSL_variable_finder::GLSL_variable_finder(std::string programText,GLUI *glui)
{
	std::istringstream prog(programText);
	std::string line;
	while (getline(prog,line)) {
		std::istringstream linestream(line);
		std::string kind,type,name; 
		linestream>>kind>>type>>name;
		if (kind=="uniform" || kind=="const") {
			size_t loc;
			loc=name.find(";");
			if (loc!=std::string::npos) name=name.substr(0,loc);
			loc=name.find("=");
			if (loc!=std::string::npos) name=name.substr(0,loc);
		
			std::cout<<"Read "<<kind<<" variable '"<<name<<"', a "<<type<<"\n";
			float value=0.0;
			loc=line.find("="); /* start of assignment */
			if (loc!=std::string::npos) { /* has an assignment */
				std::string strvalue=line.substr(loc+1);
				std::istringstream streamvalue(strvalue);
				streamvalue>>value;
			}
			GLSL_variable *u=new GLSL_variable;
			u->name=name;
			u->location=-1; /* meaning unknown & unknowable */
			u->value=value;
			variables[name]=u; /* add to our map */
			
			if (kind=="uniform" && type=="float")
			{
				u->location=-2; /* meaning "please look up" */
				new GLUI_Spinner(glui,name.c_str(),&u->value);
			}
		}
	}
}

std::string pixelShader;
GLSL_variable_finder *vf;

int stop=0; /* if true, stop running the simulation */
int limit_framerate=1; /* if true, clamp the framerate to 100Hz */

void make_ui(GLUI *glui,const char *shaderFile)
{
	new GLUI_Checkbox(glui,"Stop Simulation",&stop);
	new GLUI_Checkbox(glui,"Limit Framerate",&limit_framerate);
	new GLUI_Spinner(glui,"Display Alpha",&display_alpha);
	
	pixelShader=readFileIntoString(shaderFile);
	vf=new GLSL_variable_finder(pixelShader,glui);
}



/* Return the wall clock time, in seconds */
double oglTimer(void) {return 0.001*glutGet(GLUT_ELAPSED_TIME);}
/* Check if an error has happened--if so, exit */
void oglCheck(const char *where) {
	int err=0;
	if (0!=(err=glGetError())) {
		fprintf(stderr,"OpenGL error %s at %s\n",gluErrorString(err),where);
		exit(1);
	}
}

/* Make an OpenGL texture handle, of this size, with this pixel format */
GLuint make_3d_texture(int w,int h,int d,GLenum format=GL_RGBA8)
{
	GLuint tex=0;
	glGenTextures(1,&tex);
	GLenum t=GL_TEXTURE_3D;
	glBindTexture(t,tex);
	glTexImage3D(t,0,format, w,h,d, 0, GL_RGBA,GL_UNSIGNED_BYTE,0);
	glTexParameteri(t, GL_TEXTURE_MAG_FILTER,magfilter_mode); 
	glTexParameteri(t, GL_TEXTURE_MIN_FILTER,minfilter_mode); /* no mipmaps */
	glTexParameteri(t, GL_TEXTURE_WRAP_S, wrap_mode); 
	glTexParameteri(t, GL_TEXTURE_WRAP_T, wrap_mode);
	glTexParameteri(t, GL_TEXTURE_WRAP_R, wrap_mode);
	glGenerateMipmapEXT(t); /* make mipmaps */
	glBindTexture(t,0); /* back to default texture */
	oglCheck("after make_3d_texture");
	return tex;
}

/* Attach the current framebuffer to this z-level of this 3D texture */
void attach_3d_texture(GLuint voxel_tex,int level)
{
	glFramebufferTexture3DEXT(GL_FRAMEBUFFER_EXT, 
		GL_COLOR_ATTACHMENT0_EXT, GL_TEXTURE_3D, 
		voxel_tex, 0, level);
}

/* Make a framebuffer object that renders pixels into this texture */
GLuint make_fbo(GLuint voxel_tex,int level)
{
	GLuint fb=0;
	glGenFramebuffersEXT(1, &fb);
	glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, fb);
	attach_3d_texture(voxel_tex,level);
	
	/* Make sure the framebuffer's happy */
	GLenum status = glCheckFramebufferStatusEXT(GL_FRAMEBUFFER_EXT);
	if (status!=GL_FRAMEBUFFER_COMPLETE_EXT)
	{ fprintf(stderr,"Error %d in framebuffer creation!\n",status); exit(1); }

	glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, 0); /* back to screen render */
	oglCheck("after make_fbo");
	return fb;
}

// Call glVertex, and glTexCoords using this matrix
void vertex_with_texture(const vec3 &w)
{
	glTexCoord3fv(w);
	glVertex3fv(w);
}

void display(void) /* Draw everything in our window */
{
	oglCheck("before display function");
	static double elapsed=0.0; /* last elapsed display time (seconds) */
	static float display_dt=0.0;
	float big=300; /* size of the ground */
	float floor=0.0; /* Z height of floor */
	glFinish(); double start_t=oglTimer();
	
	glEnable(GL_BLEND);
	glBlendFunc(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA);
	
	
/* Do offscreen rendering work */

	/* This matrix projects the last frame's world coordinates into texture coords */
	static osl::mat4 toTexCoords(1.0);

	/* load an image file directly as a new OpenGL texture */
	GLenum texture_format=GL_RGBA8; // or try GL_RGBA_FLOAT16_ATI;
	int w=128; /* <-- voxel dimensions of our 3D texture */
	int h=w, d=w;
	static GLuint destTex = make_3d_texture(w,h,d,texture_format);
	static GLuint srcTex = make_3d_texture(w,h,d,texture_format);
	static GLuint fb = make_fbo(destTex,0);
	
  if (!stop) {
  
	/* Render into 3D texture using framebuffer object */
	glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, fb);
	glViewport(0,0,w,h);
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	oglCheck("After offscreen rendering matrix setup");

	glDisable(GL_BLEND);
	glDisable(GL_DEPTH_TEST);

	/* Actual physics happens right on the graphics card */
	static GLhandleARB prog = makeProgramObject(
	"varying vec3 texCoords;\n"
	"void main(void) /* vertex shader */ {\n"
	"	vec4 mv = gl_ModelViewProjectionMatrix * gl_Vertex; \n"
	"	texCoords = vec3(gl_Vertex); \n"
	"	gl_Position = mv;\n"
	"}\n"
	,
	pixelShader.c_str()
	);
	glUseProgramObjectARB(prog);
	glUniform1iARB(glGetUniformLocationARB(prog,"srcTex"),0); /* texture unit 0 */
	vf->setUniforms(prog);
	glColor4f(1,1,1,1); 
	
	glTranslatef(-1.0,-1.0,-1.0);
	glScalef(2.0,2.0,2.0);
	
	/* Render each slice of the simulation */
	for (int depth=0;depth<d;depth++) {
		attach_3d_texture(destTex,depth);

		/* run our program over the whole rectangle of the texture */
		glBegin(GL_QUADS);
		float depthz=depth/(d-1.0f); /* slice Z, scaled to 0.0-1.0 texture coordinates */
		glVertex3f( 0.0f, 0.0f,depthz);
		glVertex3f(+1.0f, 0.0f,depthz);
		glVertex3f(+1.0f,+1.0f,depthz);
		glVertex3f( 0.0f,+1.0f,depthz);
		glEnd();
	}
	oglCheck("After running simulation");
	glUseProgramObjectARB(0);
	
	if (fmod(start_t,2.0)<1.0) { /* interact with one slice of simulation */
		attach_3d_texture(destTex,d/2);
		glPushMatrix();
		glDisable(GL_TEXTURE_2D);
		glDisable(GL_TEXTURE_3D);
		glTranslatef(0.5,0.5,0.5);
		glColor4f(1.0,1.0,1.0,1.0);
		glutSolidTeapot(0.1);
		glPopMatrix();
	}

	if (limit_framerate) {
	#ifdef _WIN32
		Sleep(10);
	#else
		usleep(10*1000); 
	#endif	
	}
	std::swap(srcTex,destTex); /* ping-pong the buffers */
	
	oglCheck("After offscreen work");
  } /* end if (!stop) */
  
/* Main camera setup */
	glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, 0);
	glViewport(0,0,glutGet(GLUT_WINDOW_WIDTH),glutGet(GLUT_WINDOW_HEIGHT));
	glClearColor(0.3,0.5,0.7,1.0); /* clear to a nice blue color */
	glClear(GL_COLOR_BUFFER_BIT|GL_DEPTH_BUFFER_BIT);
	
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	
	/* Set up the main camera projection matrix */
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	gluPerspective(70.0, /* degree FOV */
		glutGet(GLUT_WINDOW_WIDTH)/float(glutGet(GLUT_WINDOW_HEIGHT)),
		0.1,1000.0 /* Z range */
	);
	oglCameraLookAt(0.5);

	
// Draw in our ground plane texture
	GLenum t=GL_TEXTURE_3D;
	glBindTexture(t,srcTex);
	glEnable(t);
	glGenerateMipmapEXT(t); /* make mipmaps before rendering */
	
	glTexParameteri(t, GL_TEXTURE_MAG_FILTER,magfilter_mode); 
	glTexParameteri(t, GL_TEXTURE_MIN_FILTER,minfilter_mode);
	glTexParameteri(t, GL_TEXTURE_WRAP_S, wrap_mode); 
	glTexParameteri(t, GL_TEXTURE_WRAP_T, wrap_mode);
	glTexParameteri(t, GL_TEXTURE_MAX_ANISOTROPY_EXT,texaniso?16:1);
	
	glEnable(GL_BLEND);
	glBegin(GL_QUADS);
	glColor4f(1,1,1,display_alpha); 
	/* Draw the whole 3D texture as a set of slices */
	for (int depth=0;depth<d;depth++) {
		/* run our program over the whole rectangle of the texture */
		float depthz=depth/(d-1.0f); /* slice Z, scaled to 0.0-1.0 texture coordinates */
		glTexCoord3f(0.0f,0.0f,depthz); glVertex3f(0.0f,0.0f,depthz);
		glTexCoord3f(1.0f,0.0f,depthz); glVertex3f(1.0f,0.0f,depthz);
		glTexCoord3f(1.0f,1.0f,depthz); glVertex3f(1.0f,1.0f,depthz);
		glTexCoord3f(0.0f,1.0f,depthz); glVertex3f(0.0f,1.0f,depthz);
	}
	glEnd();
	glDisable(t);
	
	oglCameraAxes();
	
	glutSwapBuffers(); /* display everything we've rendered so far */
	glutPostRedisplay(); /* continual animation */
	
	// Print frames per second
	glFinish(); elapsed=oglTimer()-start_t; /* stop the stopwatch */
	static double last_print=0.0; /* Periodically print out the time */
	if (start_t>last_print+0.1) {
		display_dt=elapsed+0.001;
		printf("%.1f frames/sec, %.2f ms/frame\n",
			1.0/elapsed,1000.0*elapsed);
		last_print=start_t;
	}
	oglCheck("after display function");
}

int main(int argc,char **argv) {
	glutInit(&argc,argv);  /* set up GLUT */
	
	const char *shaderFile="shader.txt";
	if (argc>1) shaderFile=argv[1]; /* first argument: shader to run */
	
	glutInitWindowPosition(0,60);
	glutInitWindowSize(1000,700);
	glutInitDisplayMode(GLUT_DOUBLE | GLUT_RGBA | GLUT_DEPTH);
	int win=glutCreateWindow("Physics From 'shader.txt'"); /* make a window */
	glewInit(); /* only AFTER create window! */
	
	// Make some popups & sliders (this is the only GLUI code you'll need!)
        GLUI *glui = GLUI_Master.create_glui("Parameters",0,0,0);
        make_ui(glui,shaderFile);
	
	glutDisplayFunc(display); /* here's how to draw the window */
	oglCameraInit();
	//camera=vec3(0.5,0.5,0.80);
	//camera_orient.y=vec3(0,1,0);camera_orient.z=vec3(0,0,1);
	correct_head_tilt=false;
	camera=vec3(0.5,0.5,1.90);
	camera_orient.x=vec3(1,0,0);
	camera_orient.y=vec3(0,1,0);
	camera_orient.z=vec3(0,0,0.3);
	glutMainLoop(); /* start drawing */
}
